// # rainbow-dash
// Very fast color interpolation in Javascript
var dashCore = require('../out/dash-core');

// ## calculateM
// Given two points on a line, calculate `m` for `y = mx + b`
var calculateM = dashCore.cwrap('calculateM', 'number', ['number', 'number', 'number', 'number']);

// ## calculateB
// Given two points on a line, calculate `b` for `y = mx + b`
var calculateB = dashCore.cwrap('calculateB', 'number', ['number', 'number', 'number', 'number']);

// ## rgbStringToTriple
// Converts the CSS `rgb(r, g, b)` form into an array of three integers
function rgbStringToTriple(rgbString) {
    return rgbString.match(/[0-9]+/g).map(Number);
}

// ## tripleToRgbString
// Converts three integers with values between [0, 255] to the CSS `rgb(r, g, b)` form
function tripleToRgbString(r, g, b) {
    return 'rgb(' + r + ',' + g + ',' + b + ')';
}

// ## hexStringToTriple
// Converts the CSS `#RRGGBB` form into an array of three integers
function hexStringToTriple(hexString) {
    var color = parseInt(hexString.substring(1), 16);
    
    var red = Math.floor(color / (256*256));
    var green = Math.floor(color / 256) % 256;
    var blue = color % (256);
    
    return [red, green, blue];
}

// ## tripleToHashString
// Converts three integers with values between [0, 255] to the CSS `#RRGGBB` form
function tripleToHashString(r, g, b) {
    return '#' + (1*256*256*256 + r*256*256 + g*256 + b).toString(16).substring(1);
}

// ## stringToTriple
// Determines which kind of CSS string it's dealing with and parses it into an array of three integers
function stringToTriple(string) {
    if(string[0] === 'r') return rgbStringToTriple(string);
    if(string[0] === '#') return hexStringToTriple(string);
    throw new Error('Unknown CSS color string');
}

// ## rainbowDash
// Given a set of stop points and their corresponding colors, returns a method that
// calculates the correct color for any arbitrary value
module.exports = function rainbowDash(options, stops) {
    // Determine if the optional options object has been provided, or just color stops
    if(!stops) {
        stops = options;
        options = {};
    }

    // Validates the options (if any) and set up rainbow-dash to use them, valid options are:
    // `outputFormat` - A string with a value of `rgb` or `hex`; default `rgb`
    // `inputRange` - An array of two numbers specifying the input range allowed; overrides `minX` and `maxX` if provided
    // `uniformStops` - A boolean indicating whether or not the stops are uniformly spaced; if so, `stops` must be an array of colors, default: `false`
    switch(options.outputFormat) {
    case 'rgb':
        /* falls through */
    case 'hex':
        break;
    default:
        options.outputFormat = 'rgb';
        break;
    }
    var tripleToString = options.outputFormat === 'rgb' ? tripleToRgbString : tripleToHashString;
    if(options.inputRange) {
        if(!(typeof(options.inputRange) === 'object' && options.inputRange instanceof Array)) {
            delete options.inputRange;
        } else if(options.inputRange.length !== 2 || typeof(options.inputRange[0]) !== 'number' || typeof(options.inputRange[1]) !== 'number') {
            delete options.inputRange;
        }
    }
    if(options.uniformStops && typeof(stops) === 'object' && stops instanceof Array && options.inputRange) {
        var newStops = {};
        var numStops = stops.length;
        var stopDelta = (options.inputRange[1] - options.inputRange[0]) / (numStops - 1);
        for(var k = 0; k < numStops; k++) {
            newStops[stopDelta*k+ options.inputRange[0]] = stops[k];
        }
        stops = newStops;
    }

    // Take the stop points, make sure they're numbers (object keys are always strings) and store them in a sorted array
    var sortedStops = Object.keys(stops).map(Number).sort(function(a, b) { return a - b; });
    var stopLen = sortedStops.length;
    var bufSortedStops = dashCore._malloc(8*stopLen); // 8 bytes for a 64-bit double
    dashCore.HEAPF64.set(sortedStops, bufSortedStops/8);

    // Store the min and max stop points and colors for fast bounds checking
    var minX = options.inputRange ? options.inputRange[0] : sortedStops[0];
    var maxX = options.inputRange ? options.inputRange[1] : sortedStops[sortedStops.length-1];
    var minY = tripleToString.apply(this, stringToTriple(stops[minX]));
    var maxY = tripleToString.apply(this, stringToTriple(stops[maxX]));

    // The loop iterates through each range (one less than the total number of stops)
    // and calculates `m` and `b` for each range and each color channel in that range
    // These are then stored into three double typed arrays holding pairs of M,B values
    // This is for optimal representation when passing to the `getTriple` function
    var rConstants = [], gConstants = [], bConstants = [];
    for(var i = 0; i < stopLen-1; i++) {
        var startX = sortedStops[i];
        var endX = sortedStops[i+1];
        var startYs = stringToTriple(stops[startX]);
        var endYs = stringToTriple(stops[endX]);
        rConstants[2*i] = calculateM(startX, endX, startYs[0], endYs[0]);
        rConstants[2*i+1] = calculateB(startX, endX, startYs[0], endYs[0]);
        gConstants[2*i] = calculateM(startX, endX, startYs[1], endYs[1]);
        gConstants[2*i+1] = calculateB(startX, endX, startYs[1], endYs[1]);
        bConstants[2*i] = calculateM(startX, endX, startYs[2], endYs[2]);
        bConstants[2*i+1] = calculateB(startX, endX, startYs[2], endYs[2]);
    }
    var bufRConstants = dashCore._malloc(8*2*(stopLen-1));
    dashCore.HEAPF64.set(rConstants, bufRConstants/8);
    var bufGConstants = dashCore._malloc(8*2*(stopLen-1));
    dashCore.HEAPF64.set(gConstants, bufGConstants/8);
    var bufBConstants = dashCore._malloc(8*2*(stopLen-1));
    dashCore.HEAPF64.set(bConstants, bufBConstants/8);

    // ### calculateRGB
    // This is the returned function after initialization; it performs a range check
    // that short-circuits for any values beyond the stop points. If within range,
    // it iterates through the sorted stop points to find the interpolation constants
    // it needs. A simple while loop is used because the number of stop points will
    // most likely be low, and a binary search will probably gain little due to the
    // calculation overhead involved. Once the correct constants are found, the input
    // value is interpolated for all three channels, and the result is stringified
    // into the CSS `rgb(r, g, b)` format.
    return function calculateRGB(x) {
        if(x <= minX) return minY;
        if(x >= maxX) return maxY;
        var triplePointer = dashCore._getTriple(bufSortedStops, bufRConstants, bufGConstants, bufBConstants, x);
        return tripleToString(
            dashCore.HEAPU8[triplePointer],
            dashCore.HEAPU8[triplePointer+1],
            dashCore.HEAPU8[triplePointer+2]
        );
    };
};
